/*
 * Tiny TPC-B 1.0 - data loader
 *
 * [Oracle Database]
 * shell> sqlplus "/ AS SYSDBA"
 * sql> CREATE USER tpcb IDENTIFIED BY tpcb;
 * sql> GRANT connect, resource TO tpcb;
 *
 * [MySQL]
 * shell> mysql -u root [-p]
 * sql> CREATE DATABASE tpcb;
 * sql> GRANT all privileges ON tpcb.* TO tpcb@'%' IDENTIFIED BY 'tpcb';
 *
 * [PostgreSQL]
 * shell> psql -U postgres
 * sql> CREATE DATABASE tpcb;
 * sql> CREATE USER tpcb PASSWORD 'tpcb';
 *
 * <postgresql.conf>
 * listen_addresses = '*'
 * port = 5432
 *
 * <pg_hba.conf>
 * host all all 0.0.0.0/0 md5
 */

// JdbcRunner settings -----------------------------------------------

// Oracle Database
// var jdbcDriver = "oracle.jdbc.OracleDriver";
// var jdbcUrl = "jdbc:oracle:thin://@localhost:1521/ORCL";

// MySQL
var jdbcDriver = "com.mysql.jdbc.Driver";
var jdbcUrl = "jdbc:mysql://localhost:3306/tpcb?rewriteBatchedStatements=true";

// PostgreSQL
// var jdbcDriver = "org.postgresql.Driver";
// var jdbcUrl = "jdbc:postgresql://localhost:5432/tpcb";

var jdbcUser = "tpcb";
var jdbcPass = "tpcb";
var isLoad = true;
var isAutoCommit = false;
var logDir = "logs";

// Application settings ----------------------------------------------

var TID_SCALE = 10;
var AID_SCALE = 100000;
var BATCH_SIZE = 100;

// JdbcRunner functions ----------------------------------------------

function init() {
    if (getId() == 0) {
        var scale = param0;
        var taskQueue = new java.util.concurrent.LinkedBlockingQueue();

        info("Tiny TPC-B 1.0.0 - data loader");
        info("-param0  : Scale factor (default : 1)");
        info("-nAgents : Parallel loading degree (default : 1)");

        if (scale == 0) {
            scale = 1;
        }
        info("Scale factor            : " + scale);
        info("Parallel loading degree : " + nAgents);

        for (var bid = 1; bid <= scale; bid++) {
            taskQueue.offer(bid);
        }
        putData("TaskQueue", taskQueue);

        if (getDatabaseProductName() == "Oracle") {
            dropTable();
            createTableOracle();
        } else if (getDatabaseProductName() == "MySQL") {
            dropTable();
            createTableMySQL();
        } else if (getDatabaseProductName() == "PostgreSQL") {
            dropTable();
            createTablePostgreSQL();
        } else {
            error(getDatabaseProductName() + " is not supported yet.");
        }

        commit();
    }
}

function run() {
    var bid = Number(getData("TaskQueue").poll());

    if (bid != 0) {
        loadData(bid);
        commit();
    } else {
        setBreak();
    }
}

function fin() {
    if (getId() == 0) {
        if (getDatabaseProductName() == "Oracle") {
            createIndexOracle();
            gatherStatsOracle();
        } else if (getDatabaseProductName() == "MySQL") {
            // NOOP
        } else if (getDatabaseProductName() == "PostgreSQL") {
            createIndexPostgreSQL();
            gatherStatsPostgreSQL();
        } else {
            error(getDatabaseProductName() + " is not supported yet.");
        }

        commit();

        info("Completed.");
        info("branches : " + fetchAsArray("SELECT COUNT(*) FROM branches")[0][0]);
        info("tellers  : " + fetchAsArray("SELECT COUNT(*) FROM tellers")[0][0]);
        info("accounts : " + fetchAsArray("SELECT COUNT(*) FROM accounts")[0][0]);
        info("history  : " + fetchAsArray("SELECT COUNT(*) FROM history")[0][0]);
    }
}

// Application functions ---------------------------------------------

function dropTable() {
    info("Dropping tables ...");

    dropTableByName("history");
    dropTableByName("accounts");
    dropTableByName("tellers");
    dropTableByName("branches");
}

function dropTableByName(tableName) {
    try {
        execute("DROP TABLE " + tableName);
    } catch (e) {
        warn(e);
        rollback(); // PostgreSQL requires a rollback
    }
}

function createTableOracle() {
    info("Creating tables ...");

    execute("CREATE TABLE branches ("
        + "bid NUMBER, "
        + "bbalance NUMBER, "
        + "filler CHAR(88))");

    execute("CREATE TABLE tellers ("
        + "tid NUMBER, "
        + "bid NUMBER, "
        + "tbalance NUMBER, "
        + "filler CHAR(84))");

    execute("CREATE TABLE accounts ("
        + "aid NUMBER, "
        + "bid NUMBER, "
        + "abalance NUMBER, "
        + "filler CHAR(84))");

    execute("CREATE TABLE history ("
        + "tid NUMBER, "
        + "bid NUMBER, "
        + "aid NUMBER, "
        + "delta NUMBER, "
        + "mtime TIMESTAMP, "
        + "filler CHAR(22))");
}

function createTableMySQL() {
    info("Creating tables ...");

    execute("CREATE TABLE branches ("
        + "bid INT PRIMARY KEY, "
        + "bbalance INT, "
        + "filler CHAR(88)) "
        + "ENGINE = InnoDB");

    execute("CREATE TABLE tellers ("
        + "tid INT PRIMARY KEY, "
        + "bid INT, "
        + "tbalance INT, "
        + "filler CHAR(84)) "
        + "ENGINE = InnoDB");

    execute("CREATE TABLE accounts ("
        + "aid INT PRIMARY KEY, "
        + "bid INT, "
        + "abalance INT, "
        + "filler CHAR(84)) "
        + "ENGINE = InnoDB");

    execute("CREATE TABLE history ("
        + "id INT PRIMARY KEY AUTO_INCREMENT, " // A surrogate key for ascending inserts
        + "tid INT, "
        + "bid INT, "
        + "aid INT, "
        + "delta INT, "
        + "mtime DATETIME, "
        + "filler CHAR(22)) "
        + "ENGINE = InnoDB");
}

function createTablePostgreSQL() {
    info("Creating tables ...");

    execute("CREATE TABLE branches ("
        + "bid INT, "
        + "bbalance INT, "
        + "filler CHAR(88))");

    execute("CREATE TABLE tellers ("
        + "tid INT, "
        + "bid INT, "
        + "tbalance INT, "
        + "filler CHAR(84))");

    execute("CREATE TABLE accounts ("
        + "aid INT, "
        + "bid INT, "
        + "abalance INT, "
        + "filler CHAR(84))");

    execute("CREATE TABLE history ("
        + "tid INT, "
        + "bid INT, "
        + "aid INT, "
        + "delta INT, "
        + "mtime TIMESTAMP, "
        + "filler CHAR(22))");
}

function createIndexOracle() {
    info("Creating indexes ...");

    execute("ALTER TABLE branches ADD CONSTRAINT pk_branches PRIMARY KEY (bid)");
    execute("ALTER TABLE tellers ADD CONSTRAINT pk_tellers PRIMARY KEY (tid)");
    execute("ALTER TABLE accounts ADD CONSTRAINT pk_accounts PRIMARY KEY (aid)");
}

function createIndexPostgreSQL() {
    info("Creating indexes ...");

    execute("ALTER TABLE branches ADD CONSTRAINT pk_branches PRIMARY KEY (bid)");
    execute("ALTER TABLE tellers ADD CONSTRAINT pk_tellers PRIMARY KEY (tid)");
    execute("ALTER TABLE accounts ADD CONSTRAINT pk_accounts PRIMARY KEY (aid)");
}

function gatherStatsOracle() {
    info("Analyzing tables ...");

    execute("BEGIN DBMS_STATS.GATHER_SCHEMA_STATS(ownname => NULL); END;");
}

function gatherStatsPostgreSQL() {
    info("Vacuuming and analyzing tables ...");

    takeConnection().setAutoCommit(true);
    execute("VACUUM ANALYZE branches");
    execute("VACUUM ANALYZE tellers");
    execute("VACUUM ANALYZE accounts");
    execute("VACUUM ANALYZE history");
}

function loadData(bid) {
    info("Loading bid " + bid + " by agent " + getId() + " ...");

    // branches
    execute("INSERT INTO branches (bid, bbalance, filler) "
        + "VALUES ($int, 0, "
        + "'bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb"
        + "bbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbbb')",
        bid);

    // tellers
    var tellersTidArray = new Array(TID_SCALE);
    var tellersBidArray = new Array(TID_SCALE);

    for (var i = 0; i < TID_SCALE; i++) {
        tellersTidArray[i] = TID_SCALE * (bid - 1) + 1 + i;
        tellersBidArray[i] = bid;
    }

    executeBatch("INSERT INTO tellers (tid, bid, tbalance, filler) "
        + "VALUES ($int, $int, 0, "
        + "'tttttttttttttttttttttttttttttttttttttttttt"
        + "tttttttttttttttttttttttttttttttttttttttttt')",
        tellersTidArray, tellersBidArray);

    // accounts
    var accountsAidArray = new Array(BATCH_SIZE);
    var accountsBidArray = new Array(BATCH_SIZE);

    for (var i = 0; i < AID_SCALE / BATCH_SIZE; i++) {
        for (var j = 0; j < BATCH_SIZE; j++) {
            accountsAidArray[j] = AID_SCALE * (bid - 1) + 1 + BATCH_SIZE * i + j;
            accountsBidArray[j] = bid;
        }

        executeBatch("INSERT INTO accounts (aid, bid, abalance, filler) "
            + "VALUES ($int, $int, 0, "
            + "'ccccccccccccccccccccccccccccccccccccccccc"
            + "ccccccccccccccccccccccccccccccccccccccccccc')",
            accountsAidArray, accountsBidArray);
    }
}
