package jdbcrunner;

import static org.junit.Assert.*;

import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;

import jdbcrunner.Template.DataType;

import org.junit.Test;
import org.mozilla.javascript.Context;
import org.mozilla.javascript.ContextFactory;
import org.mozilla.javascript.ImporterTopLevel;
import org.mozilla.javascript.Scriptable;
import org.mozilla.javascript.WrappedException;

public class TemplateTest {

	@Test
	public void testTemplate() {
		Template template;

		template = new Template("SELECT ename FROM emp");
		assertEquals("SELECT ename FROM emp", template.getPreparableStatement());
		assertTrue(template.getDataTypeList().size() == 0);

		template = new Template("SELECT ename FROM emp WHERE empno = $int");
		assertEquals("SELECT ename FROM emp WHERE empno = ?", template.getPreparableStatement());
		assertTrue(template.getDataTypeList().size() == 1);
		assertTrue(template.getDataTypeList().get(0).equals(DataType.INT));

		template = new Template("SELECT ename FROM emp WHERE ename = $string");
		assertEquals("SELECT ename FROM emp WHERE ename = ?", template.getPreparableStatement());
		assertTrue(template.getDataTypeList().size() == 1);
		assertTrue(template.getDataTypeList().get(0).equals(DataType.STRING));

		template = new Template("SELECT ename FROM emp WHERE empno = $int and ename = $string");
		assertEquals("SELECT ename FROM emp WHERE empno = ? and ename = ?", template
				.getPreparableStatement());
		assertTrue(template.getDataTypeList().size() == 2);
		assertTrue(template.getDataTypeList().get(0).equals(DataType.INT));
		assertTrue(template.getDataTypeList().get(1).equals(DataType.STRING));

		template = new Template("SELECT ename FROM emp WHERE ename = 'scott$$tiger'");
		assertEquals("SELECT ename FROM emp WHERE ename = 'scott$tiger'", template
				.getPreparableStatement());
		assertTrue(template.getDataTypeList().size() == 0);

		try {
			template = new Template(
					"SELECT ename FROM emp WHERE empno = $error and ename = $string");
			fail("IllegalArgumentExceptionがスローされませんでした");
		} catch (IllegalArgumentException e) {
			assertEquals(Resources.getString("Template.ILLEGAL_PARAMETER_NAME")
					+ "$error and ename = ", e.getMessage());
		}

		try {
			template = new Template("SELECT ename FROM emp WHERE ename = 'scott$tiger'");
			fail("IllegalArgumentExceptionがスローされませんでした");
		} catch (IllegalArgumentException e) {
			assertEquals(Resources.getString("Template.ILLEGAL_PARAMETER_NAME") + "$tiger'", e
					.getMessage());
		}

		// FindBugs用のダミー処理
		template.getPreparableStatement();
	}

	@Test
	public void testSetParameters() {
		Connection connection = null;
		Statement statement = null;
		PreparedStatement preparedStatement = null;
		ResultSet resultSet = null;

		try {
			connection = DriverManager.getConnection("jdbc:mysql://localhost/jdbcrunner",
					"jdbcrunner", "jdbcrunner");
			connection.setAutoCommit(false);

			statement = connection.createStatement();
			statement.execute("DROP TABLE IF EXISTS test01");
			statement
					.execute("CREATE TABLE test01 (id INT PRIMARY KEY, data1 int, data2 varchar(100))");
			statement.execute("INSERT INTO test01 (id, data1, data2) VALUES (1, 100, 'abc')");
			statement.execute("INSERT INTO test01 (id, data1, data2) VALUES (2, 200, 'def')");
			statement.execute("INSERT INTO test01 (id, data1, data2) VALUES (3, 300, 'ghi')");
			connection.commit();

			Context context = ContextFactory.getGlobal().enterContext();
			Scriptable scope = new ImporterTopLevel(context);
			context.evaluateString(scope, "function setParameters() "
					+ "{ template.setParameters(preparedStatement, Array.slice(arguments, 0)); }",
					"", 1, null);

			Template template = null;
			template = new Template("SELECT id, data1, data2 FROM test01 WHERE id = $int");
			preparedStatement = connection.prepareStatement(template.getPreparableStatement());
			scope.put("template", scope, template);
			scope.put("preparedStatement", scope, preparedStatement);

			context.evaluateString(scope, "setParameters(0);", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(1);", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertTrue(resultSet.next());
			assertEquals("abc", resultSet.getString(3));
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(5 / 2);", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertTrue(resultSet.next());
			assertEquals("def", resultSet.getString(3));
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(\"03\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertTrue(resultSet.next());
			assertEquals("ghi", resultSet.getString(3));
			assertFalse(resultSet.next());
			resultSet.close();

			try {
				context.evaluateString(scope, "setParameters(\"a\");", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_STRING_FOR_INT_1") + "1"
						+ Resources.getString("Template.ILLEGAL_STRING_FOR_INT_2") + "a", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(new java.util.Date());", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_1") + "1"
						+ Resources.getString("Template.ILLEGAL_NUMBER_2") + "java.util.Date", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters();", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "0"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "1", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(1, 2);", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "2"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "1", we
						.getMessage());
			}

			preparedStatement.close();

			template = new Template("SELECT id, data1, data2 FROM test01 WHERE data2 = $string");
			preparedStatement = connection.prepareStatement(template.getPreparableStatement());
			scope.put("template", scope, template);
			scope.put("preparedStatement", scope, preparedStatement);

			context.evaluateString(scope, "setParameters(\"xyz\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(\"abc\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertTrue(resultSet.next());
			assertEquals(100, resultSet.getInt(2));
			assertFalse(resultSet.next());
			resultSet.close();

			try {
				context.evaluateString(scope, "setParameters(1);", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_STRING_1") + "1"
						+ Resources.getString("Template.ILLEGAL_STRING_2") + "java.lang.Double", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters();", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "0"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "1", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(\"def\", \"ghi\");", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "2"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "1", we
						.getMessage());
			}

			preparedStatement.close();

			template = new Template(
					"SELECT id, data1, data2 FROM test01 WHERE data1 = $int and data2 = $string");
			preparedStatement = connection.prepareStatement(template.getPreparableStatement());
			scope.put("template", scope, template);
			scope.put("preparedStatement", scope, preparedStatement);

			context.evaluateString(scope, "setParameters(100, \"xyz\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(0, \"abc\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertFalse(resultSet.next());
			resultSet.close();

			context.evaluateString(scope, "setParameters(100, \"abc\");", "", 1, null);
			resultSet = preparedStatement.executeQuery();
			assertTrue(resultSet.next());
			assertEquals(1, resultSet.getInt(1));
			assertFalse(resultSet.next());
			resultSet.close();

			try {
				context.evaluateString(scope, "setParameters(\"abc\", \"abc\");", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_STRING_FOR_INT_1") + "1"
						+ Resources.getString("Template.ILLEGAL_STRING_FOR_INT_2") + "abc", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(100, 100);", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_STRING_1") + "2"
						+ Resources.getString("Template.ILLEGAL_STRING_2") + "java.lang.Double", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(100, new java.util.Date());", "", 1,
						null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_STRING_1") + "2"
						+ Resources.getString("Template.ILLEGAL_STRING_2") + "java.util.Date", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(100);", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "1"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "2", we
						.getMessage());
			}

			try {
				context.evaluateString(scope, "setParameters(100, \"abc\", \"def\");", "", 1, null);
				fail("WrappedExceptionがスローされませんでした");
			} catch (WrappedException e) {
				Throwable we = e.getWrappedException();
				assertEquals(Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_1") + "3"
						+ Resources.getString("Template.ILLEGAL_NUMBER_OF_PARAMETERS_2") + "2", we
						.getMessage());
			}

		} catch (SQLException e) {
			fail(e.getMessage());

		} finally {
			if (resultSet != null) {
				try {
					resultSet.close();
				} catch (SQLException e) {
					// NOOP
				}
			}

			if (preparedStatement != null) {
				try {
					preparedStatement.close();
				} catch (SQLException e) {
					// NOOP
				}
			}

			if (statement != null) {
				try {
					statement.close();
				} catch (SQLException e) {
					// NOOP
				}
			}

			if (connection != null) {
				try {
					connection.close();
				} catch (SQLException e) {
					// NOOP
				}
			}

			try {
				Context.exit();
			} catch (IllegalStateException e) {
				// NOOP
			}
		}
	}
}
